""""
推特工具类
"""
import datetime
import re
import requests
import time
import urllib3
from loguru import logger
from math import ceil
from my_fake_useragent import UserAgent
from urllib.parse import unquote, urljoin, quote
from easy_spider_tool import jsonpath, cookie_to_dic, clear_value, regx_match, for_to_regx_match, date_parse
from requests import RequestException

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from easy_twitter_crawler.twitter_crawler.twitter_types import *

bearer = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'


def get_proxy():
    """
    获取代理
    """
    proxy_meta = {
        'http': 'http://127.0.0.1:10808',
        'https': 'http://127.0.0.1:10808'
    }
    return proxy_meta


def get_headers(cookie: str):
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36 Edg/112.0.1722.68',
        'content-type': 'application/json',
        'x-twitter-active-user': 'yes',
        # 'x-twitter-auth-type': 'OAuth2Session',
        'authorization': bearer,
    }

    # 无需登录
    if not cookie:
        headers.update({
            'x-guest-token': gen_guest_token()
        })
    else:
        headers.update({
            'x-csrf-token': cookie_to_dic(cookie)['ct0'],
            'cookie': cookie,
        })
    return headers


def extract_time(date_obj: str) -> str:
    """
    时间处理
    """
    if not date_obj:
        return ''

    date_obj = (date_parse(
        date_obj.replace('+0000', ''),
        is_fmt=False,
        default_time_scheme=None,
        # timezone='Asia/Shanghai'
    ) + datetime.timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")

    return date_obj


def extract_error_info(json_data: Dict) -> PartialType:
    """
    处理错误消息
    @param json_data:
    @return:
    """

    return {
        'msg': jsonpath(
            json_data,
            [
                '$.errors..message',
                # 推文不可用/推文需要登录
                '$..itemContent.tweet_results.result.tombstone..text.text',
                '$..text.text',
                # 用户报错信息
                '$.data.user.result..text',
            ],
            first=True,
            default=''
        ),
        'name': jsonpath(
            json_data,
            '$.errors..name',
            first=True,
            default=''
        ),
        'type': jsonpath(
            json_data,
            '$.errors..kind',
            first=True,
            default=''
        ),
        'code': jsonpath(
            json_data,
            '$.errors..code',
            first=True,
            default=''
        ),
    }


def extract_cursor(json_data: Dict) -> str:
    """
    处理翻页cursor
    @param json_data:
    @return:
    """
    expr = [
        # 文章评论
        "$.data..instructions[0].entries.[?(@.cursorType=='Bottom'||@.cursorType=='ShowMoreThreads'||@.cursorType=='ShowMoreThreadsPrompt')].value",
        # 用户粉丝/关注
        "$.data.user.result.timeline.timeline.instructions[*].entries.[?(@.cursorType=='Bottom')].value",
        # 用户文章
        "$..entries.[?(@.cursorType=='Bottom')].value",
        # 搜索
        "$.data.search_by_raw_query.search_timeline.timeline.instructions..[?(@.cursorType=='Bottom')].value"
    ]

    cursor = jsonpath(
        json_data,
        expr,
        first=True,
        default=''
    )
    return cursor


def generating_page_link(base_url: str, cursor: str = None, **kwargs) -> Optional[str]:
    """
    根据cursor生成链接
    """
    page_url = None
    article_id = kwargs.pop('article_id', False)
    user_id = kwargs.pop('user_id', False)
    key_word = kwargs.pop('key_word', False)

    # 文章页
    if article_id:
        page_url = base_url.format(article_id=article_id, cursor=cursor)
    # 用户页
    if user_id:
        page_url = base_url.format(user_id=user_id, cursor=cursor)
    # 搜索页
    if key_word:
        data_type = kwargs.pop('data_type', '')
        page_url = base_url.format(key_word=key_word, data_type=data_type, cursor=cursor)

    return page_url


def generating_page_links(base_url, total_num=50, limit=20):
    """
    根据总数及每页显示的个数生成下一页的urls
    @param base_url: 基础的url
    @param total_num: 数据总数 默认50
    @param limit: 每页展示的个数 默认20
    """
    page_urls = []
    for i in range(ceil(total_num / limit)):
        page_urls.append(re.sub(rf'start=\d+&count=\d+', f'start={i * limit}&count={limit}', base_url))
    return page_urls


def find_element(element, entity_id: str, key: str = 'entryId', first=True, default=None) -> Union[
    List[Dict[str, any]], Dict[str, any]]:
    if default is None:
        default = '' if first is True else []
    return jsonpath(element, f'$.[?(@["{key}"]=="{entity_id}")]', first=first, default=default)


def get_useragent():
    """
    随机获取useragent
    Returns: useragent
    """
    return UserAgent(phone=False, family='chrome', os_family='windows').random()


def gen_guest_token():
    """获取guest_token参数"""
    url_base = f'https://api.twitter.com/1.1/guest/activate.json'
    headers_ua = {
        'authorization': bearer,
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36',
    }
    response = request('post', url_base, headers=headers_ua, proxies=get_proxy())
    if response is None:
        return
    res_json = response.json()
    guest_token = res_json.get('guest_token')
    return guest_token


class MaxRequestRetryCount(Exception):
    pass


def request(method: str, url: str, headers: Dict[str, str] = None, retry_num: int = 3, **kwargs) -> Response:
    retry_count = 0
    while retry_num >= retry_count:
        try:
            return requests.request(method.lower(), url, headers=headers, verify=False, **kwargs)

        except RequestException as rex:
            logger.error(
                f'request exception: {rex}' + f', kwargs: {kwargs}, next retry ...' if retry_count < retry_num else 'exit retry')
            retry_count += 1

    raise MaxRequestRetryCount(f'retry count: {retry_count}')



"""
自定义异常类
"""
from enum import Enum


class ErrorCode(Enum):
    """异常错误码"""
    pass


class AccountErrorCode(ErrorCode):
    pass


class ProxyException(Exception):
    """
    代理异常
    """
    pass


class CaptchaException(Exception):
    """
    验证异常
    """
    pass


class UserException(Exception):
    """
    用户异常信息
    """

    def __init__(self, user, msg: str = ''):
        self.user = user
        self.msg = msg

    def __str__(self):
        return f'【{self.user}】: {self.msg}'


class UserLoginException(UserException):
    """用户登录异常"""

    def __str__(self):
        return f'【{self.user}】 : 该用户信息采集需要登录'


class UserLoseException(UserException):
    """账户不存在"""

    def __str__(self):
        return f'【{self.user}】: 该账号不存在'


class UserPrivateException(UserException):
    """用户私密"""

    def __str__(self):
        return f'【{self.user}】 : 该用户已设置为私密，无法采集'


class UserSuspendedException(UserException):
    """用户封禁"""

    def __str__(self):
        return f'【{self.user}】 : 该用户已被封禁，无法采集'


class ArticleException(Exception):
    """文章异常信息"""

    def __init__(self, article_id: str, msg: str = ''):
        self.article_id = article_id
        self.msg = msg

    def __str__(self):
        return f'【{self.article_id}】: {self.msg}'


class ArticleLoginException(ArticleException):
    """文章登录异常"""

    def __str__(self):
        return f'【{self.article_id}】: 该推文采集需要登录'



# 推特基础URL
BASE_URL = 'https://twitter.com/'
# 推特基础api
BASE_API = 'https://twitter.com/i/api/'

# 推特搜索页基础api (热门：Top 用户：People 最新：Latest 视频：Videos 照片：Photos)
# SEARCH_URL = 'https://twitter.com/i/api/graphql/GcjM7tlxA-EAM98COHsYwg/SearchTimeline?variables=%7B%22rawQuery%22%3A%22{key_word}%22%2C%22count%22%3A20%2C%22cursor%22%3A%22{cursor}%22%2C%22querySource%22%3A%22typed_query%22%2C%22product%22%3A%22{data_type}%22%7D%26features%3D%7B%22rweb_lists_timeline_redesign_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'
SEARCH_URL = 'https://twitter.com/i/api/graphql/GcjM7tlxA-EAM98COHsYwg/SearchTimeline?variables=%7B%22rawQuery%22%3A%22{key_word}%22%2C%22count%22%3A40%2C%22cursor%22%3A%22{cursor}%22%2C%22querySource%22%3A%22typed_query%22%2C%22product%22%3A%22{data_type}%22%7D&features=%7B%22rweb_lists_timeline_redesign_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'

# 用户url
USER_BASE_URL = 'https://twitter.com/{user_name}'

# 个人基础信息URL
USER_BASE_API = 'https://twitter.com/i/api/graphql/rePnxwe9LZ51nQ7Sn_xN_A/UserByScreenName?variables=%7B%22screen_name%22%3A%22{user_name}%22%2C%22withSafetyModeUserFields%22%3Atrue%2C%22withSuperFollowsUserFields%22%3Atrue%7D&features=%7B%22responsive_web_twitter_blue_verified_badge_is_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Afalse%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%7D'
# 用户推文
USER_ARTICLE_URL = 'https://api.twitter.com/graphql/sj-BEQ0Jq5AwrydqFstdvg/UserTweets?variables=%7B%22userId%22%3A%22{user_id}%22%2C%22count%22%3A40%2C%22{cursor}includePromotedContent%22%3Atrue%2C%22withQuickPromoteEligibilityTweetFields%22%3Atrue%2C%22withSuperFollowsUserFields%22%3Atrue%2C%22withDownvotePerspective%22%3Afalse%2C%22withReactionsMetadata%22%3Afalse%2C%22withReactionsPerspective%22%3Afalse%2C%22withSuperFollowsTweetFields%22%3Atrue%2C%22withVoice%22%3Atrue%2C%22withV2Timeline%22%3Atrue%7D&features=%7B%22responsive_web_twitter_blue_verified_badge_is_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22vibe_api_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Afalse%2C%22interactive_text_enabled%22%3Atrue%2C%22responsive_web_text_conversations_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'
# 用户回复
USER_REPLY_URL = 'https://twitter.com/i/api/graphql/1-5o8Qhfc2kWlu_2rWNcug/UserTweetsAndReplies?variables=%7B%22userId%22%3A%22{user_id}%22%2C%22count%22%3A20%2C%22cursor%22%3A%22{cursor}%22%2C%22includePromotedContent%22%3Atrue%2C%22withCommunity%22%3Atrue%2C%22withVoice%22%3Atrue%2C%22withV2Timeline%22%3Atrue%7D&features=%7B%22rweb_lists_timeline_redesign_enabled%22%3Afalse%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D&fieldToggles=%7B%22withAuxiliaryUserLabels%22%3Afalse%2C%22withArticleRichContentState%22%3Afalse%7D'
# 用户媒体
USER_MEDIA_URL = ''
# 用户喜欢
USER_LIKES_URL = ''
# 用户粉丝
USER_FOLLOWERS_API = 'https://api.twitter.com/graphql/6AB4F2ntV7x0Oad-FvRHbw/Followers?variables=%7B%22userId%22%3A%22{user_id}%22%2C%22count%22%3A20%2C{cursor}%22includePromotedContent%22%3Afalse%2C%22withSuperFollowsUserFields%22%3Atrue%2C%22withDownvotePerspective%22%3Afalse%2C%22withReactionsMetadata%22%3Afalse%2C%22withReactionsPerspective%22%3Afalse%2C%22withSuperFollowsTweetFields%22%3Atrue%7D&features=%7B%22responsive_web_twitter_blue_verified_badge_is_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Afalse%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22vibe_api_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22freedom_of_speech_not_reach_appeal_label_enabled%22%3Afalse%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Afalse%2C%22interactive_text_enabled%22%3Atrue%2C%22responsive_web_text_conversations_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'
# 用户关注
USER_FOLLOWING_API = 'https://api.twitter.com/graphql/n1HqDGhrfjoOv2VUOtjbtg/Following?variables=%7B%22userId%22%3A%22{user_id}%22%2C%22count%22%3A20%2C{cursor}%22includePromotedContent%22%3Afalse%2C%22withSuperFollowsUserFields%22%3Atrue%2C%22withDownvotePerspective%22%3Afalse%2C%22withReactionsMetadata%22%3Afalse%2C%22withReactionsPerspective%22%3Afalse%2C%22withSuperFollowsTweetFields%22%3Atrue%7D&features=%7B%22responsive_web_twitter_blue_verified_badge_is_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Afalse%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22vibe_api_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22freedom_of_speech_not_reach_appeal_label_enabled%22%3Afalse%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Afalse%2C%22interactive_text_enabled%22%3Atrue%2C%22responsive_web_text_conversations_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'

# 文章URL
ARTICLE_URL = 'https://twitter.com/{user_name}/status/{article_id}'
ARTICLE_API = 'https://twitter.com/i/api/graphql/2ICDjqPd81tulZcYrtpTuQ/TweetResultByRestId?variables=%7B%22tweetId%22%3A%22{article_id}%22%2C%22withCommunity%22%3Afalse%2C%22includePromotedContent%22%3Afalse%2C%22withVoice%22%3Afalse%7D&features=%7B%22creator_subscriptions_tweet_preview_api_enabled%22%3Atrue%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Atrue%2C%22responsive_web_twitter_article_tweet_consumption_enabled%22%3Afalse%2C%22tweet_awards_web_tipping_enabled%22%3Afalse%2C%22freedom_of_speech_not_reach_fetch_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Atrue%2C%22longform_notetweets_rich_text_read_enabled%22%3Atrue%2C%22longform_notetweets_inline_media_enabled%22%3Atrue%2C%22responsive_web_graphql_exclude_directive_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_media_download_video_enabled%22%3Afalse%2C%22responsive_web_graphql_skip_user_profile_image_extensions_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'

# 评论API
COMMENT_API = 'https://api.twitter.com/graphql/6lWNh96EXDJCXl05SAtn_g/TweetDetail?variables=%7B%22focalTweetId%22%3A%22{article_id}%22%2C%22{cursor}with_rux_injections%22%3Atrue%2C%22includePromotedContent%22%3Atrue%2C%22withCommunity%22%3Atrue%2C%22withQuickPromoteEligibilityTweetFields%22%3Atrue%2C%22withBirdwatchNotes%22%3Afalse%2C%22withSuperFollowsUserFields%22%3Atrue%2C%22withDownvotePerspective%22%3Afalse%2C%22withReactionsMetadata%22%3Afalse%2C%22withReactionsPerspective%22%3Afalse%2C%22withSuperFollowsTweetFields%22%3Atrue%2C%22withVoice%22%3Atrue%2C%22withV2Timeline%22%3Atrue%7D&features=%7B%22responsive_web_twitter_blue_verified_badge_is_enabled%22%3Atrue%2C%22verified_phone_label_enabled%22%3Afalse%2C%22responsive_web_graphql_timeline_navigation_enabled%22%3Atrue%2C%22longform_notetweets_consumption_enabled%22%3Afalse%2C%22tweetypie_unmention_optimization_enabled%22%3Atrue%2C%22vibe_api_enabled%22%3Atrue%2C%22responsive_web_edit_tweet_api_enabled%22%3Atrue%2C%22graphql_is_translatable_rweb_tweet_is_translatable_enabled%22%3Atrue%2C%22view_counts_everywhere_api_enabled%22%3Atrue%2C%22standardized_nudges_misinfo%22%3Atrue%2C%22tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled%22%3Afalse%2C%22interactive_text_enabled%22%3Atrue%2C%22responsive_web_text_conversations_enabled%22%3Afalse%2C%22responsive_web_enhance_cards_enabled%22%3Afalse%7D'

#  CURSOR拼接规则
JOIN_CURSOR = 'cursor":"{cursor}","referrer":"tweet","'
USER_ARTICLE_CURSOR = 'cursor":"{cursor}==","'
USER_REPLY_CURSOR = '{cursor}'
USER_FOLLOWERS_CURSOR = '"cursor":"{cursor}",'
USER_FOLLOWING_CURSOR = USER_FOLLOWERS_CURSOR

#  文章类型 转推/引推/原发(None)
RETWEETED = 'retweet'
QUOTED = 'quoted'

# 数据类型
ARTICLE = 'article'
REPLY = 'reply'
USER = 'user'
COMMENT = 'comment'
FOLLOWERS = 'followers'
FOLLOWING = 'following'

# 搜索类型
TOP = 'Top'
PEOPLE = 'People'
LATEST = 'Latest'
VIDEOS = 'Videos'
PHOTOS = 'Photos'

ALL_SEARCH_TYPE = [
    TOP,
    PEOPLE,
    LATEST,
    VIDEOS,
    PHOTOS
]

# 协程数
ASYNC_COUNT = 1

#  翻页数
DEFAULT_PAGE_LIMIT = 30

# 采集评论数
DEFAULT_COMMENT_COUNT = 50

# 默认请求休眠时间
DEFAULT_REQUESTS_TIMEOUT = 10

# 默认请求等待时间
DEFAULT_REQUESTS_SLEEP = 0



from .twitter_types import *
from .twitter_scraper import TwitterScraper
from .constants import *

_scraper = TwitterScraper()


def set_cookie(cookie=None):
    """
    :param cookie: cookie 字典；格式为{'d_c0': '.....'}；
    """
    if cookie is None:
        cookie = {}

    cookies = []
    for key in cookie.keys():
        value = cookie[key]
        cookie_str = f'{key}={value}'
        cookies.append(cookie_str)
    _scraper.default_headers['cookie'] = '; '.join(cookies)


def set_proxy(proxy: Optional[Dict[str, str]] = None):
    """
    设置代理；每次请求都切换一次代理才算最佳
    """
    _scraper.set_proxy(proxy)


def _set_timeout(timeout: int):
    assert isinstance(timeout, int), 'timeout值应为大于0的整数'
    if timeout < 5 or timeout < 0:
        timeout = DEFAULT_REQUESTS_TIMEOUT
    _scraper.set_timeout(timeout=timeout)


def search_crawler(key_word: Optional[str] = None,
                   count: Optional[int] = 0,
                   comment_count: Optional[int] = -1,
                   **kwargs) -> Union[Iterator[UserType],
Iterator[ArticleType],
Iterator[VideoType],
Iterator[PhotoType]]:
    """
    关键词搜索采集（...）
    :param key_word: 关键词
    :param kwargs:
    :param count: (int) 采集指定数量的。该值过大可能会导致多次请求。-1 不采集 0采集全部（默认） >0采集指定的数量
    :param comment_count: (int) 采集指定数量的评论。该值过大可能会导致多次请求。默认-1 不采集 0采集全部 >0采集指定的数量
    @ page_limit (int): 需要采集的页数, 默认为constants下的 DEFAULT_PAGE_LIMIT
    @ data_type:（str or list or set or tuple） 获取数据类型 可选择（热门：Top 用户：People 最新：Latest 视频：Videos 照片：Photos） 大小写均可，默认所有类型都会采集
    @ drill_down_count: (int) 下钻内容采集数量(问题下的回答)，，默认-1 不采集 0采集全部 >0采集指定的数量
    :return:
    """
    _set_timeout(kwargs.pop('timeout', DEFAULT_REQUESTS_TIMEOUT))
    _scraper.requests_kwargs['timeout'] = kwargs.pop('timeout', DEFAULT_REQUESTS_TIMEOUT)
    options: Union[Dict[str, Any], Set[str]] = kwargs.setdefault('options', {})
    if isinstance(options, set):
        options = {k: True for k in options}
    options.setdefault('key_word', key_word)
    options['drill_down_count'] = kwargs.pop('drill_down_count', -1)
    options['comment_count'] = comment_count
    data_types = kwargs.get('data_type', ALL_SEARCH_TYPE)
    kwargs['data_type'] = options['data_type'] = data_types
    kwargs['sort'] = options['sort'] = kwargs.get('sort', None)
    kwargs['time_interval'] = options['time_interval'] = kwargs.get('time_interval', None)
    kwargs['total_count'] = count

    data_types = [data_types] if isinstance(data_types, str) else data_types
    if isinstance(data_types, list):
        for data_type in data_types:
            kwargs['data_type'] = data_type.capitalize()
            for result in _scraper.search_crawler(key_word, **kwargs):
                yield result


def common_crawler(task_id: Union[str],
                   data_type: Optional[str] = None,
                   drill_down_count: Optional[int] = -1,
                   comment_count: Optional[int] = -1,
                   **kwargs) -> Union[Iterator[CommentType],
Iterator[ArticleType]]:
    """
    通用采集()
    :param task_id:
    :param data_type: 指定数据类型的采集 (article or comment)
    :param drill_down_count: (int) 下钻内容采集数量 (文章可下钻采集评论)，默认-1 不采集 0采集全部 >0采集指定的数量
    :param comment_count: (int) 采集指定数量的评论。该值过大可能会导致多次请求;默认-1 不采集 0采集全部 >0采集指定的数量
    :return:
    """
    _set_timeout(kwargs.pop('timeout', DEFAULT_REQUESTS_TIMEOUT))
    options: Union[Dict[str, Any], Set[str]] = kwargs.setdefault('options', {})
    if isinstance(options, set):
        options = {k: True for k in options}
    if data_type:
        options['data_type'] = data_type
    kwargs['pubdate_sort'] = kwargs.get('pubdate_sort', True)
    options['drill_down_count'] = kwargs['drill_down_count'] = drill_down_count
    options['comment_count'] = comment_count
    options['task_id'] = task_id
    options['start_time'] = kwargs.get('start_time', '')
    options['end_time'] = kwargs.get('end_time', '')
    kwargs['total_count'] = comment_count

    assert data_type in (ARTICLE, COMMENT), '匹配不到可以采集的数据类型，请校对data_type的值'

    if data_type == ARTICLE:
        return _scraper.article_crawler(article_id=task_id, **kwargs)
    elif data_type == COMMENT:
        return _scraper.comment_crawler(article_id=task_id, **kwargs)


def user_crawler(user_id: Union[str],
                 **kwargs) -> Iterator[UserType]:
    """
    账号采集
    :param user_id: (str) 账号id 如 https://twitter.com/phivolcs_dost中 phivolcs_dost中为user_id
                    数据api：

    @ followers_count： 是否采集该账号的关注列表(粉丝) -1不采集 默认。0采集全部（可能会导致多次请求）；>0将采集指定数量
    @ following_count： 是否采集该账号的好友列表（关注别人） -1不采集 默认。0采集全部（可能会导致多次请求）；>0将采集指定数量
    @ article_count： 是否采集该账号的文章列表 -1不采集 默认。0采集全部（可能会导致多次请求）；>0将采集指定数量
    @ drill_down_count: 是否向下钻取内容（）-1不采集 默认。0采集全部（可能会导致多次请求）；>0将采集指定数量
    """
    _set_timeout(kwargs.pop('timeout', DEFAULT_REQUESTS_TIMEOUT))
    options: Union[Dict[str, Any], Set[str]] = kwargs.setdefault('options', {})
    if isinstance(options, set):
        options = {k: True for k in options}
    options['followers_count'] = kwargs.pop('followers_count', -1)
    options['following_count'] = kwargs.pop('following_count', -1)
    options['article_count'] = kwargs.pop('article_count', -1)
    options['reply_count'] = kwargs.pop('reply_count', -1)
    options['drill_down_count'] = kwargs.pop('drill_down_count', -1)
    kwargs['sort'] = options['sort'] = kwargs.get('sort', 'created')
    options['user_id'] = user_id
    options['start_time'] = kwargs.get('start_time', '')
    options['end_time'] = kwargs.get('end_time', '')

    return _scraper.user_crawler(user_id, **kwargs)



from typing import Any, Callable, Dict, Iterable, Tuple, List, Union, Optional, Set, Iterator
from requests_html import Element
from requests import Response

URL = str
Options = Dict[str, Any]

Keyword = Dict[str, List]
CommentType = Dict[str, Any]

# 搜索类型
UserType = Dict[str, Any]
ArticleType = Dict[str, Any]
VideoType = Dict[str, Any]
PhotoType = Dict[str, Any]

#
RawPage = Element
RawPost = Element
Page = Iterable[RawPost]
Credentials = Tuple[str, str]
RequestFunction = Callable[[URL], Response]
PartialType = Optional[Dict[str, Any]]



from setuptools import setup, find_packages

setup(
    name='easy_twitter_crawler',
    version='1.0.01',
    packages=find_packages(),
    license='MIT',
    author='hanxinkong',
    author_email='xinkonghan@gmail.com',
    description='简易、强大的推特（Twitter）采集程序,支持用户,发文,评论等采集',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    install_requires=[
        # List your package's dependencies here
        "easy_spider_tool>=1.0.7",
        "loguru>=0.7.0",
        "my_fake_useragent>=0.2.1",
        "requests>=2.27.1",
        "requests_html>=0.10.0",
        "urllib3>=1.26.15"
    ],
)



"""
页面迭代器
"""
import json
from loguru import logger

from easy_twitter_crawler.twitter_crawler.exceptions import ArticleException, UserLoseException
from easy_twitter_crawler.twitter_crawler.twitter_types import *
from easy_twitter_crawler.twitter_crawler.constants import *
from easy_twitter_crawler.twitter_utils.twitter_utils import (
    quote, jsonpath, extract_cursor, find_element,
    for_to_regx_match, regx_match, generating_page_link
)


def iter_search_pages(key_word: str, request_fn: RequestFunction, **kwargs) -> Iterator[Page]:
    """
    搜索
    :return:
    """
    start_url = kwargs.pop('start_url', None)
    data_type = kwargs.pop('data_type', '')

    kwargs.update({'key_word': key_word, 'data_type': data_type})

    if not start_url:
        key_word = key_word
        start_url = SEARCH_URL.format(key_word=key_word, data_type=data_type, cursor='')
    return generic_iter_pages(start_url, PageParser, request_fn, **kwargs)


def iter_article_pages(article_id: str, request_fn: RequestFunction, **kwargs) -> Iterator[Page]:
    start_url = kwargs.pop('start_url', None)
    if not start_url:
        start_url = ARTICLE_API.format(article_id=article_id, cursor='')
        kwargs.update({'article_id': article_id})
    return generic_iter_pages(start_url, ArticlePageParser, request_fn, **kwargs)


def iter_user_pages(user_id: str, request_fn: RequestFunction, **kwargs) -> Iterator[Page]:
    start_url = kwargs.pop('start_url', None)
    if not start_url:
        start_url = USER_BASE_API.format(user_name=f'{user_id}')
    return generic_iter_pages(start_url, UserPageParser, request_fn, **kwargs)


def iter_comment_pages(article_id: str, request_fn: RequestFunction, **kwargs) -> Iterator[Page]:
    start_url = kwargs.pop('start_url', None)
    if not start_url:
        start_url = COMMENT_API.format(article_id=article_id, cursor='')
        kwargs.update({'article_id': article_id})
    return generic_iter_pages(start_url, ArticleCommentPageParser, request_fn, **kwargs)


def generic_iter_pages(start_url, page_parser_cls, request_fn, **kwargs) -> Iterator[Page]:
    next_url = start_url

    while next_url:
        try:
            response = request_fn(next_url, **kwargs)
        except Exception as e:
            response = None
            logger.error(f'error: {e}')
        parser = page_parser_cls(response)
        page = parser.get_pages(**kwargs)
        yield page
        next_page_info = parser.get_next_page(**kwargs)
        next_url = next_page_info.get('next_url')

        if next_url:
            logger.warning(f'request next url {next_url}')
        else:
            logger.warning('last page')
            return


class PageParser:
    """
    json数据清洗
    """
    article_ele_regex = r"^tweet-.*"
    promoted_article_ele_regex = r"^promoted-tweet.*"
    user_ele_regex = r"^user-.*"

    def __init__(self, response):
        self.response = response
        self.html = None
        self.json_data = None
        self._parse()
        self.data_length = 0
        self.last_cursor = ''

    def _parse(self):
        # jsons = []
        assert self.response is not None, 'response is null'
        self.json_data = json.loads(self.response.text)
        self.html = self.response.html

    def get_raw_page(self) -> RawPage:
        return self.html

    def get_next_page(self, **kwargs) -> Dict[str, Union[str, bool]]:
        assert self.json_data is not None, 'json_data is null'
        next_cursor = extract_cursor(self.json_data)

        key_word = kwargs.pop('key_word', '')
        data_type = kwargs.pop('data_type', '')

        next_url = None

        if next_cursor and self.last_cursor != next_cursor and self.data_length > 2:
            self.last_cursor = next_cursor
            next_url = generating_page_link(SEARCH_URL, key_word=key_word, data_type=data_type, cursor=next_cursor)

        return {'next_cursor': next_cursor, 'next_url': next_url}

    def get_pages(self, **kwargs) -> Page:
        assert self.json_data is not None, 'json_data is null'
        data_list = None

        data = jsonpath(
            self.json_data,
            [
                '$.data.search_by_raw_query.search_timeline.timeline.instructions[*].entries[*]',
                # 搜索-用户
                # '$.data.search_by_raw_query.search_timeline.timeline.instructions[*].entries[*].content.itemContent.user_results.result'
            ],
            first=False,
            default=[]
        )
        self.data_length = len(data)

        matches = jsonpath(data, '$..entryId', first=False, default=[])
        # 是否为用户类型
        user_type = False
        if len(matches) > 0:
            user_type = matches[0].startswith('user-')

        matched_values = [value for value in matches if
                          for_to_regx_match(
                              [self.article_ele_regex, self.promoted_article_ele_regex, self.user_ele_regex],
                              value,
                              first=True,
                              default=False
                          )]

        data = [find_element(data, i) for i in matched_values]

        if user_type:
            data = jsonpath(data, '$..content.itemContent.user_results.result', first=False, default=[])

        if not data_list:
            data_list = data

        assert data_list is not None
        return data_list


class VideoPageParser(PageParser):
    """
    视频json数据清洗
    """
    pass


class ArticlePageParser(PageParser):
    """
    文章json数据清洗
    """

    def get_pages(self, **kwargs) -> Page:
        assert self.json_data is not None, 'json_data is null'
        data_list = None
        article_id = kwargs.pop('article_id', '')

        data = jsonpath(
            self.json_data,
            [
                f"$..instructions[0].entries[?(@.entryId=='tweet-{article_id}')]",
                "$.data.tweetResult.result"
            ],
            first=True,
            default={}
        )

        if not data:
            raise ArticleException(article_id, '推文不存在')

        reason = jsonpath(data, '$..__typename', first=True, default='')
        if 'TweetUnavailable' in reason:
            raise ArticleException(article_id, '推文不可用,可能是受保护的')

        if not data_list:
            data_list = [data]
        assert data_list is not None
        return data_list

    def get_next_page(self, **kwargs) -> Dict[str, Union[str, bool]]:
        assert self.json_data is not None, 'json_data is null'

        return {'next_cursor': '', 'next_url': ''}


class UserPageParser(PageParser):
    """
    用户json数据清洗
    """

    def get_pages(self, **kwargs) -> Page:
        assert self.json_data is not None, 'json_data is null'

        data_list = None
        data = jsonpath(
            self.json_data,
            [
                '$.data.user.result',
            ],
            first=True,
            default={}
        )

        if not data:
            # 用户不存在
            user_id = jsonpath(kwargs, '$.options.user_id', first=True, default='')
            raise UserLoseException(user=user_id)

        if not data_list:
            data_list = [data]
        assert data_list is not None
        return data_list


class UserArticlePageParser(PageParser):
    """
    用户的文章json数据清洗
    """
    pass


class ArticleCommentPageParser(PageParser):
    """
    用户的文章评论json数据清洗
    """
    cmt_ele_regex = r'^conversationthread(?!.*showmore)'
    showmore_ele_regex = r'^cursor-showmorethreads-.*'

    def get_pages(self, **kwargs) -> Page:
        assert self.json_data is not None, 'json_data is null'
        data_list = None
        comment_elements = []

        article_id = kwargs.pop('article_id', '')

        data = jsonpath(
            self.json_data,
            '$..instructions[*].entries[*]',
            first=False,
            default=[]
        )

        # -----------------提取包含评论的节点---------------------#
        article_element = find_element(data, f'tweet-{article_id}', first=False, default={})

        if bool(jsonpath(
                article_element,
                '$..itemContent.tweet_results.result.legacy.in_reply_to_status_id_str',
                first=True,
                default=''
        )):
            comment_elements.extend(article_element)

        matches = jsonpath(data, '$..entryId', first=False, default=[])
        # 是否有显示更多评论
        show_more_matched = [value for value in matches if
                             regx_match(self.showmore_ele_regex, value, first=True, default=False)]

        matched_values = [value for value in matches if
                          regx_match(self.cmt_ele_regex, value, first=True, default=False)]

        data = [find_element(data, i) for i in matched_values]
        comment_elements.extend(jsonpath(
            data,
            '$.*.content.items[*]',
            first=False,
            default=[]
        ))
        # 二次处理无效数据
        matches = jsonpath(comment_elements, '$..entryId', first=False, default=[])
        matched_values = [value for value in matches if
                          regx_match(self.cmt_ele_regex, value, first=True, default=False)]

        comment_elements = [find_element(data, i) for i in matched_values]

        if not comment_elements and not show_more_matched:
            raise ArticleException(article_id, '回复不存在')

        if not data_list:
            data_list = comment_elements
        assert data_list is not None
        return data_list

    def get_next_page(self, **kwargs) -> Dict[str, Union[str, bool]]:
        assert self.json_data is not None, 'json_data is null'
        next_cursor = extract_cursor(self.json_data)

        article_id = kwargs.pop('article_id', '')

        next_url = None
        if next_cursor and self.last_cursor != next_cursor:
            self.last_cursor = next_cursor
            join_cursor = quote(JOIN_CURSOR.format(cursor=next_cursor))
            # print(next_cursor)
            next_url = generating_page_link(COMMENT_API, article_id=article_id, cursor=join_cursor)

        return {'next_cursor': next_cursor, 'next_url': next_url}
